/*********************************************************************
 *
 * Copyright 2011 Intel Corporation
 * All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 *********************************************************************/


#include <iostream>

#include "os_string.hpp"

#include "cli_resources.hpp"

#include <fstream>

// Stolen from elfio
#define EI_NIDENT     16
#define EI_MAG0        0
#define EI_MAG1        1
#define EI_MAG2        2
#define EI_MAG3        3
#define EI_CLASS       4
#define ELFMAG0 0x7F
#define ELFMAG1  'E'
#define ELFMAG2  'L'
#define ELFMAG3  'F'
#define ELFCLASSNONE 0
#define ELFCLASS32   1
#define ELFCLASS64   2

int get_elf_class(const char* filename) {
    std::ifstream stream;
    stream.open( filename, std::ios::in | std::ios::binary );
    if ( !stream ) {
        return ELFCLASSNONE;
    }
    unsigned char e_ident[EI_NIDENT];

    // Read ELF file signature
    stream.seekg( 0 );
    stream.read( reinterpret_cast<char*>( &e_ident ), sizeof( e_ident ) );

    // Is it ELF file?
    if ( stream.gcount() != sizeof( e_ident ) ||
               e_ident[EI_MAG0] != ELFMAG0    ||
               e_ident[EI_MAG1] != ELFMAG1    ||
               e_ident[EI_MAG2] != ELFMAG2    ||
               e_ident[EI_MAG3] != ELFMAG3 ) {
        return ELFCLASSNONE;
    }
    return e_ident[EI_CLASS];
}

os::String
findProgram(const char*programFilename)
{
    os::String programPath;

    os::String processDir = os::getProcessName();
    processDir.trimFilename();

    programPath = processDir;
    programPath.join(programFilename);
    if (programPath.exists()) {
        return programPath;
    }

#ifndef _WIN32
    // Try absolute install directory
    programPath = APITRACE_PROGRAMS_INSTALL_DIR;
    programPath.join(programFilename);
    if (programPath.exists()) {
        return programPath;
    }
#endif

    return "";
}

os::String
findWrapper(const char *wrapperFilename, const char *argv0)
{
    os::String wrapperPath;

    os::String processDir = os::getProcessName();
    processDir.trimFilename();

    // Try relative build directory
    // XXX: Just make build and install directory layout match
    wrapperPath = processDir;
    wrapperPath.join("wrappers");
    wrapperPath.join(wrapperFilename);
    if (wrapperPath.exists()) {
        return wrapperPath;
    }

    // Try relative install directory
    wrapperPath = processDir;
#if defined(_WIN32)
    wrapperPath.join("..\\lib\\wrappers");
#elif defined(__APPLE__)
    wrapperPath.join("../lib/wrappers");
#else
    // If we are on a 64-bit architecture, support tracing 32-bit executables by preloading the appropriate shared library
#ifdef __LP64__
    int elfclass = get_elf_class(argv0);
    if(elfclass == ELFCLASS32) {
        wrapperPath.join("../lib/apitrace/wrappers");
        if (!wrapperPath.exists()) {
            std::cerr << "32-bit wrapper libraries not found, please install apitrace-libs.i686" << std::endl;
        }
    } else if(elfclass == ELFCLASS64) {
        wrapperPath.join("../lib64/apitrace/wrappers");
    } else if(elfclass == ELFCLASSNONE) {
        std::cerr << "error: failed to determine elf class for " << argv0 << ", assuming ELFCLASS64" << std::endl;
        wrapperPath.join("../lib64/apitrace/wrappers");
    }
#else
    wrapperPath.join("../lib/apitrace/wrappers");
#endif
#endif
    wrapperPath.join(wrapperFilename);
    if (wrapperPath.exists()) {
        return wrapperPath;
    }

#ifndef _WIN32
    // Try absolute install directory
    wrapperPath = APITRACE_WRAPPERS_INSTALL_DIR;
    wrapperPath.join(wrapperFilename);
    if (wrapperPath.exists()) {
        return wrapperPath;
    }
#endif

    return "";
}

os::String
findScript(const char *scriptFilename)
{
    os::String scriptPath;
    scriptPath = APITRACE_LIB_DIR;
    scriptPath.join("scripts");
    scriptPath.join(scriptFilename);
    if (scriptPath.exists()) {
        return scriptPath;
    }
    
    std::cerr << "error: cannot find " << scriptFilename << " script\n";

    return "";
}
